## Multimarker risk prediction and its application to
##   clinical trial enrichment using a logistic model
##
## Copyright (c) 2009-2010 - Hyun Min Kang and Youna Hu - All rights reserved.
##
## Last updated Sep 18, 2010

require(boot)

## Main function of Histogram/ROC curve estimation part
## Sample genotypes [num.geno.samples] times 
## Sample ORs [num.beta.samples] times and compute betas
SampleAndWriteHistogramAndROC <- function(params, outprefix = "CTE", num.geno.samples = 1e5, num.beta.samples = 1e2, num.hist.breaks = 25, num.roc.breaks = 1000 ) {
  ## Histogram/ROC routines using expected ORs
  ## Sample 100k sets of genotypes by using population allele frequencies
  sampled.genos <- SampleGenos(params$RAFs, params$ORs, params$preval, num.geno.samples)
  
  ## Compute risk scores beta(x), and mu
  mu.betas <- ComputeMuBetas(sampled.genos, params$ORs, params$preval)

  ## Compues Pr(x|d=0) and Pr(x|d=1) to make the summation to be 1
  beta.weights <- BetaWeights(mu.betas$mu, mu.betas$betas)
  
  ## convert to log base 10 for drawing histograms
  betas.log10 <- mu.betas$betas/log(10)
  ## Generate the weighted histogram for cases and controls
  wbhist <- WeightedBetaHistogram(betas.log10, beta.weights, breaks = num.hist.breaks)
  ## Generate ROC curve
  roc.breaks <- (0:num.roc.breaks)/num.roc.breaks
  wbroc <- WeightedBetaROC(mu.betas$betas, beta.weights, roc.breaks)

  ## Repeat the same procedure for Using very pessimistic ORs
  ## - used for ROC only
  OR.lower <- exp(log(params$ORs)-1.96*params$SE.logORs)
  sampled.genos.lower <- SampleGenos(params$RAFs, OR.lower, params$preval, num.geno.samples)
  mu.betas.lower <- ComputeMuBetas(sampled.genos.lower, OR.lower, params$preval)
  beta.weights.lower <- BetaWeights(mu.betas.lower$mu, mu.betas.lower$betas)
  wbroc.lower <- WeightedBetaROC(mu.betas.lower$betas, beta.weights.lower, roc.breaks)

  ## Using very optimistic ORs - ROC only
  OR.upper <- exp(log(params$ORs)+1.96*params$SE.logORs)
  sampled.genos.upper <- SampleGenos(params$RAFs, OR.upper, params$preval, num.geno.samples)
  mu.betas.upper <- ComputeMuBetas(sampled.genos.upper, OR.upper, params$preval)
  beta.weights.upper <- BetaWeights(mu.betas.upper$mu, mu.betas.upper$betas)
  wbroc.upper <- WeightedBetaROC(mu.betas.upper$betas, beta.weights.upper, roc.breaks)  

  ## Sample Beta randomly and get error bars and the boundary of ROC vurves
  errs <- SampledBetasStdErrors(params$RAFs, params$ORs, params$SE.logORs, params$preval, num.geno.samples, num.beta.samples, wbhist$breaks, roc.breaks)

  ## Generate the intervals of histogram
  starts <- wbhist$breaks[-length(wbhist$breaks)]
  ends <- wbhist$breaks[-1]
  mids <- (starts+ends)/2

  ## compute the mean and variance of risk scores
  RAFs.case <- CaseRAFs(params$ORs,params$RAFs)
  RAFs.cont <- params$RAFs
  mu.case <- sum(2 * RAFs.case * log(params$ORs)/log(10))
  mu.cont <- sum(2 * RAFs.cont * log(params$ORs)/log(10))
  sd.case <- sqrt(sum((log(params$ORs)/log(10))^2*2*RAFs.case*(1-RAFs.case)))
  sd.cont <- sqrt(sum((log(params$ORs)/log(10))^2*2*RAFs.cont*(1-RAFs.cont)))
  
  #mean.betas.pop <- sum(params$RAFs * log(params$ORs));
  #sd.betas <- sqrt(sum((log(params$ORs)/log(10))^2*2*params$RAFs*(1-params$RAFs)))
  #r.betas <- rnorm(1e7,0,sd.betas)
  #mu <- EstimateMu(r.betas,params$preval)
  #mu.case <- mean(1/(1+exp(0-mu-r.betas))/params$preval*(mu+r.betas))
  #mu.cont <- mean((1-1/(1+exp(0-mu-r.betas)))/(1-params$preval)*(mu+r.betas))
  #print(paste(c(mu,mu.case,mu.cont)))

  # Now (mu.case,sd.betas), (mu.cont,sd.betas) are estimated
  approx.case <- pnorm((ends-mu.case)/sd.case)-pnorm((starts-mu.case)/sd.case)
  approx.cont <- pnorm((ends-mu.cont)/sd.cont)-pnorm((starts-mu.cont)/sd.cont)

  ## Write the histogram file
  write.table(cbind(starts,ends,mids,wbhist$wprob.case,wbhist$wprob.ctrl,approx.case,approx.cont,errs$prob.mean,errs$prob.se,errs$case.mean,errs$case.se,errs$ctrl.mean,errs$ctrl.se),paste(outprefix,"whist",sep="."),row.names=F,col.names=c("start","end","mid","eOR-case","eOR-ctrl","approx-case","approx-ctrl","sampled-pop-mean","sampled-pop-se","sampled-case-mean","sampled-case-se","sampled-ctrl-mean","sampled-ctrl-se"))
  ## Write the ROC file
  write.table(cbind(roc.breaks,wbroc$roc,wbroc.lower$roc,wbroc.upper$roc,errs$roc.mean,errs$roc.se,errs$roc.min,errs$roc.max),paste(outprefix,"roc",sep="."),row.names=F,col.names=c("ctrl-breaks","expected-OR-case","very-pessimistic","very-optimistic","sampled-case-mean","sampled-case-se","roc-min","roc-max"))
  ## Write AUC information and etc.
  write.table(c(wbroc$auc,wbroc.lower$auc,wbroc.upper$auc,errs$auc.mean,errs$auc.se,params$preval),paste(outprefix,"auc",sep="."),row.names=c("AUC-Expected-OR","AUC-very-pessimistic","AUC-very-optimistic","AUC-Sample-Mean","AUC-SE","preval"),col.names=F)
}

## Main function of Enrichment trial part
## Sample genotypes [num.geno.samples] times 
## Sample ORs [num.beta.samples] times and compute betas
## Compute
##  - Sample sizes for enrichment trial design
##  - Total cost of the study
##  - Time savings
SampleAndWriteForEnrichmentTrial <- function(params, outprefix = "CTE", num.geno.samples = 1e5, num.beta.samples = 1e2, num.frac.targeted = 1000 ) {
  ## sample from a general population, and adjust mu to account for sample population difference and treatment effect
  sampled.genos <- SampleGenos(params$RAFs, params$ORs, params$preval, num.geno.samples)
  mu.betas.c <- ComputeMuBetas(sampled.genos, params$ORs, params$pi.c)
  mu.betas.t <- ComputeMuBetas(sampled.genos, params$ORs, params$pi.t)

  ## Make sure that the betas are the same between two groups
  stopifnot(mu.betas.c$betas == mu.betas.t$betas)
  mu.c <- mu.betas.c$mu
  mu.t <- mu.betas.t$mu
  betas <- mu.betas.c$betas
  rm(mu.betas.c)
  rm(mu.betas.t)

  ## Compute Pr(x|d,Rx) and Pr(x|d,Ctrl)
  beta.weights.c <- BetaWeights(mu.c, betas)
  beta.weights.t <- BetaWeights(mu.t, betas)
  frac.targeted <- (1:num.frac.targeted)/num.frac.targeted

  ## Algorithms to compute upper and lower bound of targeted propotion
  ## Compute Pr(x,d|Rx), Pr(x,d|Ctrl)
  dpreval.c <- ComputeDensityPreval(betas, beta.weights.c, params$pi.c)
  dpreval.t <- ComputeDensityPreval(betas, beta.weights.t, params$pi.t)
  prevals.c <- vector(length=num.frac.targeted)
  prevals.t <- vector(length=num.frac.targeted)
  indices <- vector(length=num.frac.targeted)
  ## For each possible targeted proportion (interval size)
  for(i in 1:num.frac.targeted) {
    ## Find the optimal interval minimizing the sample size
    r <- FindOptimalSampleSizeInterval(dpreval.c, dpreval.t, i/num.frac.targeted)
    prevals.c[i] <- r$c
    prevals.t[i] <- r$t
    indices[i] <- r$idx
  }

  ## Enrichment trial number calculation
  alpha <- 0.05  # Type I error
  beta <- 0.8    # Power
  sample.size.binary.targeted <- ComputeTargetedBinarySampleSize(prevals.c, prevals.t, alpha, beta)  # Targeted trial sample size
  sample.size.binary.standard <- sample.size.binary.targeted[length(sample.size.binary.targeted)]    # Standard trial sample size

  cost.followup.per.year <- params$cost.f
  cost.followup <- cost.followup.per.year * params$trial.years
  cost.accrual <- params$cost.s
  cost.genotyping <- params$cost.g  

  ## Screened & Pool-based trial cost
  cost.binary.targeted <- sample.size.binary.targeted * ( cost.followup + cost.accrual/frac.targeted + cost.genotyping/frac.targeted )
  cost.binary.standard <- sample.size.binary.standard * ( cost.followup + cost.accrual )
  cost.binary.pooled <- sample.size.binary.targeted * ( cost.followup + cost.accrual )

  prevals.avg <- (prevals.c+prevals.t)/2   # average incidence
  z.sq <- (qnorm(alpha/2)+qnorm(1-beta))^2 # (z_alpha+z_beta)^2
  
  prevals.c.0 <- prevals.c[length(prevals.c)] 
  prevals.t.0 <- prevals.t[length(prevals.t)]
  prevals.avg.0 <- prevals.avg[length(prevals.avg)]

  ## Pool-based time savings
  ## this computes the time savings when a=0, and the python script
  ## will account for various a/t
  time.binary.fixed.sample <- 4*prevals.avg*z.sq/(sample.size.binary.standard*(prevals.c-prevals.t)^2+4*prevals.avg^2*z.sq)

  time.standard <- rep(1,num.frac.targeted)
  cost.binary.reduced.time <- sample.size.binary.standard * ( cost.followup.per.year * time.binary.fixed.sample * params$trial.years + cost.accrual/frac.targeted + cost.genotyping/frac.targeted )
  cost.pooled.reduced.time <- sample.size.binary.standard * ( cost.followup.per.year * time.binary.fixed.sample * params$trial.years + cost.accrual )  

  write.table(cbind(frac.targeted, prevals.c, prevals.t, sample.size.binary.targeted, rep(sample.size.binary.standard,num.frac.targeted), cost.binary.targeted, cost.binary.pooled, rep(cost.binary.standard,num.frac.targeted), time.binary.fixed.sample*params$trial.years, time.standard*params$trial.years,cost.binary.reduced.time,cost.pooled.reduced.time,indices),paste(outprefix,"target",sep="."),row.names=F,col.names=c("targeted-fraction","placebo-incidence","treatment-incidence","sample-size-targeted","sample-size-standard","cost-targeted","cost-pooled","cost-standard","time-targeted-fixed-sample-size","time-standard","cost-standard-reduced-time","cost-pooled-reduced-time","optimal-indices"))
}

## hsq is heritability in liability scale
GeneralRiskEnrichmentTrial <- function(params, AUC, outprefix = "CTE", num.samples = 1e5, num.frac.targeted = 1000 ) {
  ## sample from a general population, and adjust mu to account for sample population difference and treatment effect

  r <- EstimateSigmaFromAUC(params$preval, AUC, num.samples)
  betas <- r$betas
  mu.c <- EstimateMu(betas, params$pi.c)
  mu.t <- EstimateMu(betas, params$pi.t)
                            
  beta.weights.c <- BetaWeights(mu.c, betas)
  beta.weights.t <- BetaWeights(mu.t, betas)
  
  frac.targeted <- (1:num.frac.targeted)/num.frac.targeted

  dpreval.c <- ComputeDensityPreval(betas, beta.weights.c, params$pi.c)
  dpreval.t <- ComputeDensityPreval(betas, beta.weights.t, params$pi.t)
  prevals.c <- vector(length=num.frac.targeted)
  prevals.t <- vector(length=num.frac.targeted)
  indices <- vector(length=num.frac.targeted)
  for(i in 1:num.frac.targeted) {
    r <- FindOptimalSampleSizeInterval(dpreval.c, dpreval.t, i/num.frac.targeted)
    prevals.c[i] <- r$c
    prevals.t[i] <- r$t
    indices[i] <- r$idx
  }

  alpha <- 0.05
  beta <- 0.8
  sample.size.binary.targeted <- ComputeTargetedBinarySampleSize(prevals.c, prevals.t, alpha, beta)
  sample.size.binary.standard <- sample.size.binary.targeted[length(sample.size.binary.targeted)]

  cost.followup.per.year <- params$cost.f
  cost.followup <- cost.followup.per.year * params$trial.years
  cost.accrual <- params$cost.s
  cost.genotyping <- params$cost.g

  cost.binary.targeted <- sample.size.binary.targeted * ( cost.followup + cost.accrual/frac.targeted + cost.genotyping/frac.targeted )
  cost.binary.standard <- sample.size.binary.standard * ( cost.followup + cost.accrual )
  cost.binary.pooled <- sample.size.binary.targeted * ( cost.followup + cost.accrual )

  prevals.avg <- (prevals.c+prevals.t)/2
  z.sq <- (qnorm(alpha/2)+qnorm(1-beta))^2
  
  prevals.c.0 <- prevals.c[length(prevals.c)]
  prevals.t.0 <- prevals.t[length(prevals.t)]
  prevals.avg.0 <- prevals.avg[length(prevals.avg)]

  # this computes the time savings when a=0, and the python script
  # will account for various a/t
  time.binary.fixed.sample <- 4*prevals.avg*z.sq/(sample.size.binary.standard*(prevals.c-prevals.t)^2+4*prevals.avg^2*z.sq)

  a2t <- 0.1

  time.standard <- rep(1,num.frac.targeted)
  cost.binary.reduced.time <- sample.size.binary.standard * ( cost.followup.per.year * time.binary.fixed.sample * params$trial.years + cost.accrual/frac.targeted + cost.genotyping/frac.targeted )
  cost.pooled.reduced.time <- sample.size.binary.standard * ( cost.followup.per.year * time.binary.fixed.sample * params$trial.years + cost.accrual )

  write.table(cbind(frac.targeted, prevals.c, prevals.t, sample.size.binary.targeted, rep(sample.size.binary.standard,num.frac.targeted), cost.binary.targeted, cost.binary.pooled, rep(cost.binary.standard,num.frac.targeted), time.binary.fixed.sample*params$trial.years, time.standard*params$trial.years,cost.binary.reduced.time,cost.pooled.reduced.time,indices),paste(outprefix,"target",sep="."), row.names=F, col.names=c("targeted-fraction","placebo-incidence","treatment-incidence","sample-size-targeted","sample-size-standard","cost-targeted","cost-pooled","cost-standard","time-targeted-fixed-sample-size","time-standard","cost-standard-reduced-time","cost-pooled-reduced-time","optimal-indices"))
}

## Sample genotypes randomly given RAFs and ORs i.i.d.
## Population RAFs are calculated based on expected ORs and ctrl RAFs
## Returns genotypes of (num.geno.samples * m) matrix
SampleGenos <- function(RAFs, ORs, preval, num.geno.samples) {
  ## initialize variables and sanity checking
  num.SNPs <- length(RAFs) # RAFs are control allele frequencies
  stopifnot(length(ORs) == num.SNPs)
  stopifnot( (preval > 0) && ( preval < 1 ) )

  ## calculate population risk allele frequencies
  RAFs.case <- CaseRAFs(ORs,RAFs)
  RAFs.pop <- RAFs*(1-preval) + RAFs.case*preval  # population allele frequency
  stopifnot(sum(RAFs.pop>1) == 0) # check boundary conditions
  stopifnot(sum(RAFs.pop<0) == 0)
  
  ## sample control genotypes num.geno.samples * num.SNPs matrix
  ## sample individuals bases on population allele frequencies
  sampled.genos <- matrix(rbinom(num.geno.samples * num.SNPs,2,RAFs.pop), num.geno.samples, num.SNPs, byrow=T)

  return(sampled.genos)
}

ComputeDensityPreval <- function(betas, beta.weights, preval) {
  # sanity checkes
  num.geno.samples <- length(betas)
  stopifnot(length(beta.weights$weights.case) == num.geno.samples)
  stopifnot(length(beta.weights$weights.ctrl) == num.geno.samples)

  o <- order(betas, decreasing=TRUE)

  # return Pr(d=1|x) = Pr(d=1|x)/Pr(d=1)*Pr(d=1)
  return(beta.weights$weights.case[o]*preval)
}

## Sample Beta and compute standard errors of histograms and ROC curves
SampledBetasStdErrors <- function(RAFs, ORs, SElogORs, preval, num.geno.samples, num.beta.samples, hist.breaks, roc.breaks ) {
  num.roc <- length(roc.breaks)
  num.prob <- length(hist.breaks)-1
  
  auc.sum <- 0
  auc.sumsq <- 0
  roc.sum <- rep(0,num.roc)
  roc.sumsq <- rep(0,num.roc)
  prob.sum <- rep(0,num.prob)
  prob.sumsq <- rep(0,num.prob)
  case.sum <- rep(0,num.prob)
  case.sumsq <- rep(0,num.prob)
  ctrl.sum <- rep(0,num.prob)
  ctrl.sumsq <- rep(0,num.prob)
  roc.min <- rep(1,num.roc)
  roc.max <- rep(0,num.roc)
  
  for ( i in 1:num.beta.samples ) {
    print(paste("i = ",i))
    sampled.ORs <- exp(rnorm(length(RAFs),log(ORs),SElogORs))
#    sampled.ORs <- ORs
    sampled.genos <- SampleGenos(RAFs, sampled.ORs, preval, num.geno.samples)
    mu.betas <- ComputeMuBetas(sampled.genos, sampled.ORs, preval)
    beta.weights <- BetaWeights(mu.betas$mu, mu.betas$betas)
    betas.log10 <- mu.betas$betas/log(10)
    prob <- WeightedProb(betas.log10, rep(1/num.geno.samples,num.geno.samples), hist.breaks)
    wbhist <- WeightedBetaHistogram(betas.log10, beta.weights, hist.breaks)    
    roc <- WeightedBetaROC(mu.betas$betas, beta.weights, roc.breaks)
    print(length(roc$roc))

    auc.sum <- auc.sum + roc$auc
    auc.sumsq <- auc.sumsq + roc$auc*roc$auc
    roc.sum <- roc.sum + roc$roc
    roc.sumsq <- roc.sumsq + roc$roc*roc$roc
    prob.sum <- prob.sum + prob
    prob.sumsq <- prob.sumsq + prob*prob
    case.sum <- case.sum + wbhist$wprob.case
    case.sumsq <- case.sumsq + wbhist$wprob.case * wbhist$wprob.case
    ctrl.sum <- ctrl.sum + wbhist$wprob.ctrl
    ctrl.sumsq <- ctrl.sumsq + wbhist$wprob.ctrl * wbhist$wprob.ctrl

    roc.min <- apply(cbind(roc.min,roc$roc),1,min)
    roc.max <- apply(cbind(roc.min,roc$roc),1,max)    
  }
  auc.mean <- auc.sum/num.beta.samples
  roc.mean <- roc.sum/num.beta.samples
  prob.mean <- prob.sum/num.beta.samples
  case.mean <- case.sum/num.beta.samples
  ctrl.mean <- ctrl.sum/num.beta.samples
  
  auc.se <- sqrt(auc.sumsq/num.beta.samples - auc.mean^2)
  roc.se <- sqrt(roc.sumsq/num.beta.samples - roc.mean^2)
  prob.se <- sqrt(prob.sumsq/num.beta.samples - prob.mean^2)
  case.se <- sqrt(case.sumsq/num.beta.samples - case.mean^2)
  ctrl.se <- sqrt(ctrl.sumsq/num.beta.samples - ctrl.mean^2)

  return(list(auc.mean=auc.mean,roc.mean=roc.mean,prob.mean=prob.mean,case.mean=case.mean,ctrl.mean=ctrl.mean,auc.se=auc.se,roc.se=roc.se,prob.se=prob.se,case.se=case.se,ctrl.se=ctrl.se,roc.min=roc.min,roc.max=roc.max))
}

SampleBetas <- function(sigma2, n) {
  rnorm(n,0,sqrt(sigma2))
}

ComputePreval <- function(mu, betas) {
  return (mean(1/(1+exp(0-mu-betas))))
}

## Compute weighted ROC curve using beta and weights
WeightedBetaROC <- function(betas, beta.weights, breaks.ctrl = (0:1000)/1000) {
  num.geno.samples <- length(betas)
  stopifnot(length(beta.weights$weights.case) == num.geno.samples)
  stopifnot(length(beta.weights$weights.ctrl) == num.geno.samples)

  o <- order(betas, decreasing=TRUE)
  cum.case <- c(0,cumsum(beta.weights$weights.case[o]),1)
  cum.ctrl <- c(0,cumsum(beta.weights$weights.ctrl[o]),1)

  print(system.time(breaks.case <- approx(cum.ctrl, cum.case, breaks.ctrl, ties="ordered")$y))
  print(breaks.case[length(breaks.case)])
  stopifnot(sum(is.na(breaks.case)) == 0)
  AUC <- sum((cum.case[-1]+cum.case[-(length(cum.case))])*(cum.ctrl[-1]-cum.ctrl[-(length(cum.ctrl))]))/2
  print(paste("AUC=",AUC))
  return(list(auc=AUC,roc=breaks.case,breaks=breaks.ctrl))
}

WeightedBetaAUC <- function(betas, beta.weights) {
  num.geno.samples <- length(betas)

  o <- order(betas, decreasing=TRUE)
  cum.case <- c(0,cumsum(beta.weights$weights.case[o]),1)
  cum.ctrl <- c(0,cumsum(beta.weights$weights.ctrl[o]),1)
  AUC <- sum((cum.case[-1]+cum.case[-(length(cum.case))])*(cum.ctrl[-1]-cum.ctrl[-(length(cum.ctrl))]))/2
  return(AUC)
}

ContourSigmaFromAUC <- function(mus, AUCs, num.samples = 1e5) {
  betas.n <- SampleBetas(1,num.samples)
  nmus <- length(mus)
  nAUCs <- length(AUCs)
  sigmas <- matrix(nrow=nAUCs,ncol=nmus)
  prevals <- matrix(nrow=nAUCs,ncol=nmus)

  for(i in 1:nAUCs) {
    AUC <- AUCs[i]
    for(j in 1:nmus) {
      mu <- mus[j]
      s <- uniroot(
                   function(sigma) {
                     betas <- betas.n*sigma;
                     preval <- ComputePreval(mu,betas)
                     betas.weights <- BetaWeights(mu,betas)
                     myAUC <- WeightedBetaAUC(betas,betas.weights)
                     ##print(paste(sigma,myAUC,AUC));
                     return(myAUC-AUC)
                   }, c(1e-3,1e3) )$root
      sigmas[i,j] <- s
      prevals[i,j] <- ComputePreval(mu,betas.n*s)
    }
  }
  return(list(sigmas=sigmas,prevals=prevals))
}

## Estimate the sigma based on betas, prevalence, and AUC
## sample a number of betas from standard normal, and rescale the betas to adjust the AUC values
## Return value is not deterministic, but should have relatively small variance
EstimateSigmaFromAUC <- function(preval, AUC, num.samples = 1e5) {
  betas.n <- SampleBetas(1,num.samples)
  s <- uniroot(
               function(sigma) {
                 betas <- betas.n*sigma;
                 mu <- EstimateMu(betas,preval);
                 betas.weights <- BetaWeights(mu,betas);
                 myAUC <- WeightedBetaAUC(betas,betas.weights);
                 #print(paste(sigma,myAUC,AUC));
                 return(myAUC-AUC);
               }, c(1e-3,1e3) )$root
  return(list(sigma=s,betas=betas.n*s))
}

## Computes weights on each betas conditioning on disease status
## Weight is computed as Pr(x|d)/Pr(x) = Pr(d|x)/Pr(d)
## No adjustment for difference in expected ORs and sampled ORs
BetaWeights <- function(mu, betas) {
  weights.case <- 1/(1+exp(0-(mu + betas)))
  weights.ctrl <- 1/(1+exp(mu+betas))

  # normalize the weights across all points
  weights.case <- weights.case/sum(weights.case)
  weights.ctrl <- weights.ctrl/sum(weights.ctrl)
  return(list(weights.case=weights.case,weights.ctrl=weights.ctrl))
}

## Given breaks of intervals, compute weighted counts of values
WeightedProb <- function(betas, weights, breaks) {
  m <- length(breaks)-1
  #bins <- as.integer(cut(betas,breaks=breaks,include.lowest=TRUE))
  breaks.bw <- breaks[2]-breaks[1]
  bins <- as.integer((betas-breaks[1])/breaks.bw+1.5)

  n <- length(betas)
  wprobs <- rep(0,m)
  for(i in 1:n) {
    if ( ( bins[i] > 0 ) && ( bins[i] < m+1 ) ) {
      wprobs[bins[i]] <- wprobs[bins[i]] + weights[i]
    }
  }
  return(wprobs)
}

## Function estimate.mu()
## Given a list of \beta(x), find a mu such that E[Pr(d=1|x)] = preval
EstimateMu <- function(betas, preval) {
  ## set up a start value of mu using the median of betas so that
  ## logit(preval) = mu + median(betas)
  mu.guess <- logit(preval) - median(betas)
  mu.bandwidth <- 5*sd(betas) # arbitrarily chosen bandwidth
  mu.start <- mu.guess + mu.bandwidth
  mu.end <- mu.guess - mu.bandwidth

  ## use uniroot function 
  r <- uniroot( function(mu) {
#    newpreval <- mean(1/(1+exp(0-mu-betas)))
#    print(paste(mu,preval,newpreval))
#    return(newpreval-preval)
      return(mean(1/(1+exp(0-mu-betas)))-preval)
    }, c(mu.start,mu.end) )
  mu <- return(r$root)
  return(mu)
}


## Calcualte the allele frequency for the case samples
CaseRAFs <- function(ORs, RAFs){
  ## use p+(1-p-)/p-/(1-p+) = OR
  return(ORs*RAFs/(1+(ORs-1)*RAFs))
}

# Compute per-year incidence rate based on the following equation
# pi : incidence, lambda : incidence rate, T : total years, A : accural years
# pi = 1-exp(-lambda(T-A))*(1-exp(-lambda*A))/(lambda*A)
ComputeIncidenceRate <- function(incidence, years.total, years.accural) {
  T <- years.total
  A <- years.accural
  r <- uniroot( function(x) { incidence-1+exp(0-x*(T-A))*(1-exp(0-x*A))/(x*A) }, c(1e-5,1) )
  return(r$root)
}

LoadParams <- function(name) {
  T <- read.table('all.loci.tbl')
  M <- as.matrix(T[T[,1] == name,4:7])
  RAFs <- M[,1]
  ORs <- M[,2]
  ORlowers <- M[,3]
  ORuppers <- M[,4]

  T <- read.table('all.preval.tbl')
  M <- as.double(T[T[,1] == name,2:3])
  preval <- M[1]
  sibrisk <- M[2]

  SE.logORs <- (log(ORuppers)-log(ORlowers))/3.92
  
}

InitParams <- function(diseaseName) {
  T <- read.table('all.loci.tbl')
  S <- T[T$V1 == diseaseName,]
  RAFs <- S$V4
  ORs <- S$V5
  SE.logORs <- (log(S$V7)-log(S$V6))/3.92

  T <- read.table('all.params.tbl')
  S <- T[rownames(T) == diseaseName,]
  preval <- S$preval  # population disease prevalence
  sibrisk <- S$sibrisk # sibling recurrence risk
  lambda.c <- S$lambda.c    # per-year incidence in control arm
  lambda.t <- S$lambda.t    # per-year indicence in treatment arm
  years <- S$years
  pi.c <- lambda.c * years
  pi.t <- lambda.t * years
  cost.s <- S$cost.s
  cost.g <- S$cost.g
  cost.f <- S$cost.f
  
  return(list(ORs=ORs,RAFs=RAFs,preval=preval,SE.logORs=SE.logORs,lambda.c=lambda.c,lambda.t=lambda.t,sibrisk=sibrisk,pi.c=pi.c,pi.t=pi.t,trial.years=years,cost.s=cost.s,cost.g=cost.g,cost.f=cost.f))
}

# Conversion from odds ratio to relative risk
OR.to.RR <- function( OR, RAF, preval ) {
  pcont <- RAF # RAF is control risk allele freuquency
  pcase <- OR*pcont/((OR-1)*pcont+1)
  pd <- pcase-pcont
  return(pcase*(1-pcont-preval*pd)/(1-pcase)/(pcont+preval*pd))
}
	

FindOptimalSampleSizeInterval <- function(dpreval.c, dpreval.t, frac.targeted) {
  n <- length(dpreval.c)
  stopifnot(n == length(dpreval.t))
  sz.interval <- n*frac.targeted
  num.intervals <- n-sz.interval+1
  
  cumsum.c <- c(0,cumsum(dpreval.c)) # length is n+1
  cumsum.t <- c(0,cumsum(dpreval.t))

  intsum.c <- (cumsum.c[(n-num.intervals+2):(n+1)]-cumsum.c[1:num.intervals])/frac.targeted
  intsum.t <- (cumsum.t[(n-num.intervals+2):(n+1)]-cumsum.t[1:num.intervals])/frac.targeted

  alpha = 0.05
  beta = 0.8
  sample.sizes <- ComputeTargetedBinarySampleSize(intsum.c, intsum.t, alpha, beta)
  #print(sum.c[1:10])
  idx <- which.min(sample.sizes)
  return(list(idx=idx,c=intsum.c[idx],t=intsum.t[idx]))
}

## Write weighted histograms of betas for case and controls
WriteWeightedBetaHistogram <- function(outfile, betas, beta.weights, preval, breaks = 25 ) {
  wbhist <- WeightedBetaHistogram(betas,beta.weights,preval,breaks)
  
  ## Population risk fractions - For stacked histogram
  wprobs.pop.case <- preval*wbhist$wprob.case
  wprobs.pop.ctrl <- (1-preval)*wbhist$wprob.ctrl

  num.breaks <- length(wbhist$breaks)
  starts <- wbhist$breaks[-num.breaks]
  ends <- wbhist$breaks[-1]
  mids <- (starts+ends)/2

  write.table(cbind(starts,ends,mids,wbhist$wprob.case,wbhist$wprob.ctrl,wcprobs.pop.case,wprobs.pop.ctrl),file=outfile,col.names=c("start","end","mid","Case","Control","Case","Control"),row.names=F)
}

ComputeMuBetas <- function(sampled.genos, ORs, preval) {
  #ComputeMuBetas.additive(sampled.genos,ORs,preval)
  ComputeMuBetas.multiplicative(sampled.genos,ORs,preval)
}


ComputeMuBetas.additive <- function(sampled.genos, ORs, preval) {
  ## sanity checking on the data
  num.geno.samples <- nrow(sampled.genos)
  num.SNPs <- ncol(sampled.genos)
  stopifnot(length(ORs) == num.SNPs)
  stopifnot( (preval > 0) && ( preval < 1 ) )

  #betas <- log(sampled.genos %*% (ORs-1) + 1) # nGenoSamples * 1 column vector
  #betas <- sampled.genos %*% (ORs-1) # nGenoSamples * 1 column vector
  #betas <- log(sampled.genos %*% (ORs-1) + 1)# nGenoSamples * 1 column vector   
  RRs <- ORs/(1-preval + preval*ORs)-1
  RRs.add <- sampled.genos %*% RRs + 1
  #RRs.add <- RRs.add / mean(RRs.add)
  #stopifnot(max(RRs.add*preval) < 1)
  #ORs.add <- RRs.add*(1-preval)/(1-RRs.add*preval)
  
  ## compute beta(x) for nGenoSamples
  #betas <- sampled.genos %*% log(ORs) # nGenoSamples * 1 column vector
  betas <- log(RRs.add)
  print(max(betas))
  print(min(betas))
  mu <- EstimateMu(betas,preval)

  return(list(mu=mu,betas=betas))
}

ComputeMuBetas.multiplicative <- function(sampled.genos, ORs, preval) {
  ## sanity checking on the data
  num.geno.samples <- nrow(sampled.genos)
  num.SNPs <- ncol(sampled.genos)
  stopifnot(length(ORs) == num.SNPs)
  stopifnot( (preval > 0) && ( preval < 1 ) )

  ## compute beta(x) for nGenoSamples
  betas <- sampled.genos %*% log(ORs) # nGenoSamples * 1 column vector
  mu <- EstimateMu(betas,preval)

  print(max(betas))
  print(min(betas))

  return(list(mu=mu,betas=betas))
}

## Compute weighted histograms of betas for case and controls given breaks
WeightedBetaHistogram <- function(betas.log10, beta.weights, breaks = 25 ) {
  ## sanity checking
  num.geno.samples <- length(betas.log10)

  ## use hist() function to get breaks, counts, and density
  if ( length(breaks) == 1 ) { # if breaks are given as a single number
    h <- hist(betas.log10, br=breaks, plot=FALSE) # Compute breaks
    h.breaks <- h$breaks
  }
  else {
    h.breaks <- breaks
  }
  wprob.case <- WeightedProb(betas.log10, beta.weights$weights.case, h.breaks)
  wprob.ctrl <- WeightedProb(betas.log10, beta.weights$weights.ctrl, h.breaks)

  return(list(breaks=h.breaks,wprob.case=wprob.case,wprob.ctrl=wprob.ctrl))
}

ComputeTargetedBinarySampleSize <- function(prevals.c, prevals.t, alpha, beta) {
  prevals.avg <- (prevals.c+prevals.t)/2
  return( 4*prevals.avg*(1-prevals.avg)*((qnorm(alpha/2)+qnorm(1-beta))^2)/((prevals.c-prevals.t)^2) )
}

Wray.sibrisk <- function(preval, sibrisk) {
  K <- preval
  lambda <- sibrisk
  T <- qnorm(1-K)
  z <- dnorm(T)
  T1 <- qnorm(1-lambda*K)
  i <- z/K
  hsq <- 2*(T-T1*sqrt(1-(T^2-T1^2)*(1-T/i)))/(i+T1^2*(i-T))
  v <- 0-i*K*(1-K)
  AUC <- pnorm((i-v)*hsq/sqrt(hsq*(1-hsq*i*(i-T)+1-hsq*v*(v-T))))
  return(list(hsq=hsq,AUC=AUC))
}

Wray.AUC <- function(preval, AUC) {
  K <- preval
  T <- qnorm(1-K)
  z <- dnorm(T)
  i <- z/K
  Q <- qnorm(AUC)
  v <- 0-i*K*(1-K)
  hsq <- 2*Q*Q/((v-i)^2+Q^2*i*(i-T)+v*(v-T))
  T1 <- (T-hsq*i/2)/sqrt(1-hsq^2*i*(i-T)/4)
  lambda <- (1-pnorm(T1))/K
  return(list(hsq=hsq,sibrisk=lambda))
}

Wray.hsq <- function(preval, hsq) {
  K <- preval
  T <- qnorm(1-K)
  z <- dnorm(T)
  i <- z/K
  v <- 0-i*K*(1-K)
  T1 <- (T-hsq*i/2)/sqrt(1-hsq^2*i*(i-T)/4)
  lambda <- (1-pnorm(T1))/K
  AUC <- pnorm((i-v)*hsq/sqrt(hsq*(1-hsq*i*(i-T)+1-hsq*v*(v-T))))
  return(list(AUC=AUC,sibrisk=lambda))
}
